@@ -2,6 +2,7 @@
  * sparse/show-parse.c
  *
  * Copyright (C) 2003 Transmeta Corp.
+ *               2003-2004 Linus Torvalds
  *
  *  Licensed under the Open Software License version 1.1
  *
@@ -16,6 +17,7 @@
 #include <fcntl.h>
 
 #include "lib.h"
+#include "allocate.h"
 #include "token.h"
 #include "parse.h"
 #include "symbol.h"
@@ -23,24 +25,61 @@
 #include "expression.h"
 #include "target.h"
 
+static int show_symbol_expr(struct symbol *sym);
+static int show_string_expr(struct expression *expr);
+
 static void do_debug_symbol(struct symbol *sym, int indent)
 {
 	static const char indent_string[] = "                                  ";
 	static const char *typestr[] = {
-		"base", "node", "ptr.", "fn..",
-		"arry", "strt", "unin", "enum",
-		"tdef", "tpof", "memb", "bitf",
-		"labl"
+		[SYM_UNINITIALIZED] = "none",
+		[SYM_PREPROCESSOR] = "cpp.",
+		[SYM_BASETYPE] = "base",
+		[SYM_NODE] = "node",
+		[SYM_PTR] = "ptr.",
+		[SYM_FN] = "fn..",
+		[SYM_ARRAY] = "arry",
+		[SYM_STRUCT] = "strt",
+		[SYM_UNION] = "unin",
+		[SYM_ENUM] = "enum",
+		[SYM_TYPEDEF] = "tdef",
+		[SYM_TYPEOF] = "tpof",
+		[SYM_MEMBER] = "memb",
+		[SYM_BITFIELD] = "bitf",
+		[SYM_LABEL] = "labl",
+		[SYM_RESTRICT] = "rstr",
+		[SYM_FOULED] = "foul",
+		[SYM_BAD] = "bad.",
 	};
+	struct context *context;
+	int i;
 
 	if (!sym)
 		return;
-	fprintf(stderr, "%.*s%s%3d:%lu %lx %s (as: %d, context: %x:%x) %p\n",
+	fprintf(stderr, "%.*s%s%3d:%lu %s %s (as: %d) %p (%s:%d:%d) %s\n",
 		indent, indent_string, typestr[sym->type],
 		sym->bit_size, sym->ctype.alignment,
-		sym->ctype.modifiers, show_ident(sym->ident),
-		sym->ctype.as, sym->ctype.context, sym->ctype.contextmask,
-		sym);
+		modifier_string(sym->ctype.modifiers), show_ident(sym->ident), sym->ctype.as,
+		sym, stream_name(sym->pos.stream), sym->pos.line, sym->pos.pos,
+		builtin_typename(sym) ?: "");
+	i = 0;
+	FOR_EACH_PTR(sym->ctype.contexts, context) {
+		/* FIXME: should print context expression */
+		fprintf(stderr, "< context%d: in=%d, out=%d\n",
+			i, context->in, context->out);
+		fprintf(stderr, "  end context%d >\n", i);
+		i++;
+	} END_FOR_EACH_PTR(context);
+	if (sym->type == SYM_FN) {
+		struct symbol *arg;
+		i = 0;
+		FOR_EACH_PTR(sym->arguments, arg) {
+			fprintf(stderr, "< arg%d:\n", i);
+			do_debug_symbol(arg, 0);
+			fprintf(stderr, "  end arg%d >\n", i);
+			i++;
+		} END_FOR_EACH_PTR(arg);
+	}
 	do_debug_symbol(sym->ctype.base_type, indent+2);
 }
 
@@ -60,10 +99,12 @@ const char *modifier_string(unsigned long mod)
 	const char *res,**ptr, *names[] = {
 		"auto", "register", "static", "extern",
 		"const", "volatile", "[signed]", "[unsigned]",
-		"[char]", "[short]", "[long]", "[long]",
+		"[char]", "[short]", "[long]", "[long long]",
 		"[typdef]", "[structof]", "[unionof]", "[enum]",
 		"[typeof]", "[attribute]", "inline", "[addressable]",
-		"[nocast]", "[noderef]", 
+		"[nocast]", "[noderef]", "[accessed]", "[toplevel]",
+		"[label]", "[assigned]", "[type]", "[safe]",
+		"[usertype]", "[force]", "[explicitly-signed]",
 		NULL
 	};
 	ptr = names;
@@ -80,29 +121,22 @@ const char *modifier_string(unsigned long mod)
 	return buffer;
 }
 
-void show_struct_member(struct symbol *sym, void *data, int flags)
+static void show_struct_member(struct symbol *sym)
 {
-	if (flags & ITERATE_FIRST)
-		printf(" {\n\t");
-	printf("%s:%d:%ld at offset %ld", show_ident(sym->ident), sym->bit_size, sym->ctype.alignment, sym->offset);
-	if (sym->fieldwidth)
-		printf("[%d..%d]", sym->bit_offset, sym->bit_offset+sym->fieldwidth-1);
-	if (flags & ITERATE_LAST)
-		printf("\n} ");
-	else
-		printf(", ");
-}
-
-static void show_one_symbol(struct symbol *sym, void *sep, int flags)
-{
-	show_symbol(sym);
-	if (!(flags & ITERATE_LAST))
-		printf("%s", (const char *)sep);
+	printf("\t%s:%d:%ld at offset %ld.%d", show_ident(sym->ident), sym->bit_size, sym->ctype.alignment, sym->offset, sym->bit_offset);
+	printf("\n");
 }
 
 void show_symbol_list(struct symbol_list *list, const char *sep)
 {
-	symbol_iterate(list, show_one_symbol, (void *)sep);
+	struct symbol *sym;
+	const char *prepend = "";
+
+	FOR_EACH_PTR(list, sym) {
+		puts(prepend);
+		prepend = ", ";
+		show_symbol(sym);
+	} END_FOR_EACH_PTR(sym);
 }
 
 struct type_name {
@@ -110,7 +144,7 @@ struct type_name {
 	char *end;
 };
 
-static void prepend(struct type_name *name, const char *fmt, ...)
+static void FORMAT_ATTR(2) prepend(struct type_name *name, const char *fmt, ...)
 {
 	static char buffer[512];
 	int n;
@@ -124,7 +158,7 @@ static void prepend(struct type_name *name, const char *fmt, ...)
 	memcpy(name->start, buffer, n);
 }
 
-static void append(struct type_name *name, const char *fmt, ...)
+static void FORMAT_ATTR(2) append(struct type_name *name, const char *fmt, ...)
 {
 	static char buffer[512];
 	int n;
@@ -138,45 +172,74 @@ static void append(struct type_name *name, const char *fmt, ...)
 	name->end += n;
 }
 
+static struct ctype_name {
+	struct symbol *sym;
+	const char *name;
+} typenames[] = {
+	{ & char_ctype,  "char" },
+	{ &schar_ctype,  "signed char" },
+	{ &uchar_ctype,  "unsigned char" },
+	{ & short_ctype, "short" },
+	{ &sshort_ctype, "signed short" },
+	{ &ushort_ctype, "unsigned short" },
+	{ & int_ctype,   "int" },
+	{ &sint_ctype,   "signed int" },
+	{ &uint_ctype,   "unsigned int" },
+	{ &slong_ctype,  "signed long" },
+	{ & long_ctype,  "long" },
+	{ &ulong_ctype,  "unsigned long" },
+	{ & llong_ctype, "long long" },
+	{ &sllong_ctype, "signed long long" },
+	{ &ullong_ctype, "unsigned long long" },
+
+	{ &void_ctype,   "void" },
+	{ &bool_ctype,   "bool" },
+	{ &string_ctype, "string" },
+
+	{ &float_ctype,  "float" },
+	{ &double_ctype, "double" },
+	{ &ldouble_ctype,"long double" },
+	{ &incomplete_ctype, "incomplete type" },
+	{ &int_type, "abstract int" },
+	{ &fp_type, "abstract fp" },
+	{ &label_ctype, "label type" },
+	{ &bad_ctype, "bad type" },
+};
+
+const char *builtin_typename(struct symbol *sym)
+{
+	int i;
+
+	for (i = 0; i < sizeof(typenames)/sizeof(typenames[0]); i++)
+		if (typenames[i].sym == sym)
+			return typenames[i].name;
+	return NULL;
+}
+
+const char *builtin_ctypename(struct ctype *ctype)
+{
+	int i;
+
+	for (i = 0; i < sizeof(typenames)/sizeof(typenames[0]); i++)
+		if (&typenames[i].sym->ctype == ctype)
+			return typenames[i].name;
+	return NULL;
+}
+
 static void do_show_type(struct symbol *sym, struct type_name *name)
 {
-	int i, modlen;
+	int modlen;
 	const char *mod;
-	static struct ctype_name {
-		struct symbol *sym;
-		char *name;
-	} typenames[] = {
-		{ & char_ctype,  "char" },
-		{ &uchar_ctype,  "unsigned char" },
-		{ & short_ctype, "short" },
-		{ &ushort_ctype, "unsigned short" },
-		{ & int_ctype,   "int" },
-		{ &uint_ctype,   "unsigned int" },
-		{ & long_ctype,  "long" },
-		{ &ulong_ctype,  "unsigned long" },
-		{ & llong_ctype, "long long" },
-		{ &ullong_ctype, "unsigned long long" },
-
-		{ &void_ctype,   "void" },
-		{ &bool_ctype,   "bool" },
-		{ &string_ctype, "string" },
-
-		{ &float_ctype,  "float" },
-		{ &double_ctype, "double" },
-		{ &ldouble_ctype,"long double" },
-	};
-
+	const char *typename;
 	if (!sym)
 		return;
 
-	for (i = 0; i < sizeof(typenames)/sizeof(typenames[0]); i++) {
-		if (typenames[i].sym == sym) {
-			int len = strlen(typenames[i].name);
-			*--name->start = ' ';
-			name->start -= len;
-			memcpy(name->start, typenames[i].name, len);
-			return;
-		}
+	if ((typename = builtin_typename(sym))) {
+		int len = strlen(typename);
+		*--name->start = ' ';
+		name->start -= len;
+		memcpy(name->start, typename, len);
+		return;
 	}
 
 	/* Prepend */
@@ -197,14 +260,14 @@ static void do_show_type(struct symbol *sym, struct type_name *name)
 
 	case SYM_ENUM:
 		prepend(name, "enum %s ", show_ident(sym->ident));
-		return;
+		break;
 
 	case SYM_NODE:
 		append(name, "%s", show_ident(sym->ident));
 		break;
 
 	case SYM_BITFIELD:
-		append(name, ":%d", sym->fieldwidth);
+		append(name, ":%d", sym->bit_size);
 		break;
 
 	case SYM_LABEL:
@@ -214,6 +277,12 @@ static void do_show_type(struct symbol *sym, struct type_name *name)
 	case SYM_ARRAY:
 		break;
 
+	case SYM_RESTRICT:
+		break;
+
+	case SYM_FOULED:
+		break;
+
 	default:
 		prepend(name, "unknown type %d", sym->type);
 		return;
@@ -239,8 +308,17 @@ static void do_show_type(struct symbol *sym, struct type_name *name)
 		return;
 
 	case SYM_ARRAY:
-		append(name, "[%d]", sym->array_size);
+		append(name, "[%lld]", get_expression_value(sym->array_size));
 		return;
+
+	case SYM_RESTRICT:
+		prepend(name, "restricted ");
+		return;
+
+	case SYM_FOULED:
+		prepend(name, "fouled ");
+		return;
+
 	default:
 		break;
 	}
@@ -274,7 +352,10 @@ void show_symbol(struct symbol *sym)
 
 	if (!sym)
 		return;
-		
+
+	if (sym->ctype.alignment)
+		printf(".align %ld\n", sym->ctype.alignment);
+
 	show_type(sym);
 	type = sym->ctype.base_type;
 	if (!type)
@@ -284,18 +365,29 @@ void show_symbol(struct symbol *sym)
 	 * Show actual implementation information
 	 */
 	switch (type->type) {
+		struct symbol *member;
+
 	case SYM_STRUCT:
-		symbol_iterate(type->symbol_list, show_struct_member, NULL);
-		break;
-
 	case SYM_UNION:
-		symbol_iterate(type->symbol_list, show_struct_member, NULL);
+		printf(" {\n");
+		FOR_EACH_PTR(type->symbol_list, member) {
+			show_struct_member(member);
+		} END_FOR_EACH_PTR(member);
+		printf("}\n");
 		break;
 
-	case SYM_FN:
-		printf("\n");		
-		show_statement(type->stmt);
+	case SYM_FN: {
+		struct statement *stmt = type->stmt;
+		if (stmt) {
+			int val;
+			printf("\n");		
+			val = show_statement(stmt);
+			if (val)
+				printf("\tmov.%d\t\tretval,%d\n", stmt->ret->bit_size, val);
+			printf("\tret\n");
+		}
 		break;
+	}
 
 	default:
 		break;
@@ -307,27 +399,75 @@ void show_symbol(struct symbol *sym)
 	}
 }
 
-static int show_return_stmt(struct statement *stmt)
-{
-	struct expression *expr = stmt->expression;
-
-	if (expr && expr->ctype) {
-		int val = show_expression(expr);
-		printf("\tmov.%d\t\tretval,v%d\n",
-			expr->ctype->bit_size, val);
-	}
-	printf("\tret\n");
-	return 0;
-}
-
 static int show_symbol_init(struct symbol *sym);
 
+static int new_pseudo(void)
+{
+	static int nr = 0;
+	return ++nr;
+}
+
 static int new_label(void)
 {
 	static int label = 0;
 	return ++label;
 }
 
+static void show_switch_statement(struct statement *stmt)
+{
+	int val = show_expression(stmt->switch_expression);
+	struct symbol *sym;
+	printf("\tswitch v%d\n", val);
+
+	/*
+	 * Debugging only: Check that the case list is correct
+	 * by printing it out.
+	 *
+	 * This is where a _real_ back-end would go through the
+	 * cases to decide whether to use a lookup table or a
+	 * series of comparisons etc
+	 */
+	printf("# case table:\n");
+	FOR_EACH_PTR(stmt->switch_case->symbol_list, sym) {
+		struct statement *case_stmt = sym->stmt;
+		struct expression *expr = case_stmt->case_expression;
+		struct expression *to = case_stmt->case_to;
+
+		if (!expr) {
+			printf("    default");
+		} else {
+			if (expr->type == EXPR_VALUE) {
+				printf("    case %lld", expr->value);
+				if (to) {
+					if (to->type == EXPR_VALUE) {
+						printf(" .. %lld", to->value);
+					} else {
+						printf(" .. what?");
+					}
+				}
+			} else
+				printf("    what?");
+		}
+		printf(": .L%p\n", sym->bb_target);
+	} END_FOR_EACH_PTR(sym);
+	printf("# end case table\n");
+
+	show_statement(stmt->switch_statement);
+
+	if (stmt->switch_break->used)
+		printf(".L%p:\n", stmt->switch_break->bb_target);
+}
+
+static void show_symbol_decl(struct symbol_list *syms)
+{
+	struct symbol *sym;
+	FOR_EACH_PTR(syms, sym) {
+		show_symbol_init(sym);
+	} END_FOR_EACH_PTR(sym);
+}
+
+static int show_return_stmt(struct statement *stmt);
+
 /*
  * Print out a statement
  */
@@ -336,20 +476,32 @@ int show_statement(struct statement *stmt)
 	if (!stmt)
 		return 0;
 	switch (stmt->type) {
+	case STMT_DECLARATION:
+		show_symbol_decl(stmt->declaration);
+		return 0;
 	case STMT_RETURN:
 		return show_return_stmt(stmt);
 	case STMT_COMPOUND: {
-		struct symbol *sym;
 		struct statement *s;
-		int last;
-
-		FOR_EACH_PTR(stmt->syms, sym) {
-			show_symbol_init(sym);
-		} END_FOR_EACH_PTR;
+		int last = 0;
 
+		if (stmt->inline_fn) {
+			show_statement(stmt->args);
+			printf("\tbegin_inline \t%s\n", show_ident(stmt->inline_fn->ident));
+		}
 		FOR_EACH_PTR(stmt->stmts, s) {
 			last = show_statement(s);
-		} END_FOR_EACH_PTR;
+		} END_FOR_EACH_PTR(s);
+		if (stmt->ret) {
+			int addr, bits;
+			printf(".L%p:\n", stmt->ret);
+			addr = show_symbol_expr(stmt->ret);
+			bits = stmt->ret->bit_size;
+			last = new_pseudo();
+			printf("\tld.%d\t\tv%d,[v%d]\n", bits, last, addr);
+		}
+		if (stmt->inline_fn)
+			printf("\tend_inlined\t%s\n", show_ident(stmt->inline_fn->ident));
 		return last;
 	}
 
@@ -359,6 +511,8 @@ int show_statement(struct statement *stmt)
 		int val, target;
 		struct expression *cond = stmt->if_conditional;
 
+/* This is only valid if nobody can jump into the "dead" statement */
+#if 0
 		if (cond->type == EXPR_VALUE) {
 			struct statement *s = stmt->if_true;
 			if (!cond->value)
@@ -366,6 +520,7 @@ int show_statement(struct statement *stmt)
 			show_statement(s);
 			break;
 		}
+#endif
 		val = show_expression(cond);
 		target = new_label();
 		printf("\tje\t\tv%d,.L%d\n", val, target);
@@ -380,26 +535,12 @@ int show_statement(struct statement *stmt)
 		printf(".L%d:\n", target);
 		break;
 	}
-	case STMT_SWITCH: {
-		int val = show_expression(stmt->switch_expression);
-		printf("\tswitch v%d\n", val);
-		show_statement(stmt->switch_statement);
-		if (stmt->switch_break->used)
-			printf(".L%p:\n", stmt->switch_break);
+	case STMT_SWITCH:
+		show_switch_statement(stmt);
 		break;
-	}
 
 	case STMT_CASE:
-		if (!stmt->case_expression)
-			printf("default");
-		else {
-			printf("case %lld", stmt->case_expression->value);
-			if (stmt->case_to) {
-				printf("...%lld", stmt->case_to->value);
-			}
-			
-		}
-		printf(":\n");
+		printf(".L%p:\n", stmt->case_label);
 		show_statement(stmt->case_statement);
 		break;
 
@@ -411,21 +552,24 @@ int show_statement(struct statement *stmt)
 		struct expression *post_condition = stmt->iterator_post_condition;
 		int val, loop_top = 0, loop_bottom = 0;
 
+		show_symbol_decl(stmt->iterator_syms);
 		show_statement(pre_statement);
 		if (pre_condition) {
 			if (pre_condition->type == EXPR_VALUE) {
-				if (!pre_condition->value)
-					break;
-				pre_condition = NULL;
+				if (!pre_condition->value) {
+					loop_bottom = new_label();   
+					printf("\tjmp\t\t.L%d\n", loop_bottom);
+				}
 			} else {
 				loop_bottom = new_label();
 				val = show_expression(pre_condition);
 				printf("\tje\t\tv%d, .L%d\n", val, loop_bottom);
 			}
 		}
-		if (!post_condition || post_condition->type != EXPR_VALUE || post_condition->value)
+		if (!post_condition || post_condition->type != EXPR_VALUE || post_condition->value) {
 			loop_top = new_label();
-		printf(".L%d:\n", loop_top);
+			printf(".L%d:\n", loop_top);
+		}
 		show_statement(statement);
 		if (stmt->iterator_continue->used)
 			printf(".L%p:\n", stmt->iterator_continue);
@@ -441,12 +585,11 @@ int show_statement(struct statement *stmt)
 		}
 		if (stmt->iterator_break->used)
 			printf(".L%p:\n", stmt->iterator_break);
-		if (pre_condition)
+		if (loop_bottom)
 			printf(".L%d:\n", loop_bottom);
 		break;
 	}
 	case STMT_NONE:
-		printf("\t!NONE!\n");
 		break;
 	
 	case STMT_LABEL:
@@ -457,49 +600,30 @@ int show_statement(struct statement *stmt)
 	case STMT_GOTO:
 		if (stmt->goto_expression) {
 			int val = show_expression(stmt->goto_expression);
-			printf("\tgoto *v%d\n", val);
+			printf("\tgoto\t\t*v%d\n", val);
 		} else {
-			printf("\tgoto .L%p\n", stmt->goto_label);
+			printf("\tgoto\t\t.L%p\n", stmt->goto_label->bb_target);
 		}
 		break;
 	case STMT_ASM:
 		printf("\tasm( .... )\n");
 		break;
-		
+	case STMT_CONTEXT: {
+		int val = show_expression(stmt->expression);
+		printf("\tcontext( %d )\n", val);
+		break;
+	}
+	case STMT_RANGE: {
+		int val = show_expression(stmt->range_expression);
+		int low = show_expression(stmt->range_low);
+		int high = show_expression(stmt->range_high);
+		printf("\trange( %d %d-%d)\n", val, low, high); 
+		break;
+	}	
 	}
 	return 0;
 }
 
-static void show_one_statement(struct statement *stmt, void *sep, int flags)
-{
-	show_statement(stmt);
-	if (!(flags & ITERATE_LAST))
-		printf("%s", (const char *)sep);
-}
-
-void show_statement_list(struct statement_list *stmt, const char *sep)
-{
-	statement_iterate(stmt, show_one_statement, (void *)sep);
-}
-
-static void show_one_expression(struct expression *expr, void *sep, int flags)
-{
-	show_expression(expr);
-	if (!(flags & ITERATE_LAST))
-		printf("%s", (const char *)sep);
-}
-
-void show_expression_list(struct expression_list *list, const char *sep)
-{
-	expression_iterate(list, show_one_expression, (void *)sep);
-}
-
-static int new_pseudo(void)
-{
-	static int nr = 0;
-	return ++nr;
-}
-
 static int show_call_expression(struct expression *expr)
 {
 	struct symbol *direct;
@@ -508,7 +632,7 @@ static int show_call_expression(struct expression *expr)
 	int framesize;
 
 	if (!expr->ctype) {
-		warn(expr->pos, "\tcall with no type!");
+		warning(expr->pos, "\tcall with no type!");
 		return 0;
 	}
 
@@ -518,7 +642,7 @@ static int show_call_expression(struct expression *expr)
 		int size = arg->ctype->bit_size;
 		printf("\tpush.%d\t\tv%d\n", size, new);
 		framesize += size >> 3;
-	} END_FOR_EACH_PTR;
+	} END_FOR_EACH_PTR_REVERSE(arg);
 
 	fn = expr->fn;
 
@@ -538,13 +662,19 @@ static int show_call_expression(struct expression *expr)
 		printf("\tcall\t\t*v%d\n", fncall);
 	}
 	if (framesize)
-		printf("\tadd.%d\t\tvSP,vSP,$%d\n", BITS_IN_POINTER, framesize);
+		printf("\tadd.%d\t\tvSP,vSP,$%d\n", bits_in_pointer, framesize);
 
 	retval = new_pseudo();
 	printf("\tmov.%d\t\tv%d,retval\n", expr->ctype->bit_size, retval);
 	return retval;
 }
 
+static int show_comma(struct expression *expr)
+{
+	show_expression(expr->left);
+	return show_expression(expr->right);
+}
+
 static int show_binop(struct expression *expr)
 {
 	int left = show_expression(expr->left);
@@ -554,7 +684,8 @@ static int show_binop(struct expression *expr)
 	static const char *name[] = {
 		['+'] = "add", ['-'] = "sub",
 		['*'] = "mul", ['/'] = "div",
-		['%'] = "mod"
+		['%'] = "mod", ['&'] = "and",
+		['|'] = "lor", ['^'] = "xor"
 	};
 	unsigned int op = expr->op;
 
@@ -567,6 +698,14 @@ static int show_binop(struct expression *expr)
 	return new;
 }
 
+static int show_slice(struct expression *expr)
+{
+	int target = show_expression(expr->base);
+	int new = new_pseudo();
+	printf("\tslice.%d\t\tv%d,v%d,%d\n", expr->r_nrbits, target, new, expr->r_bitpos);
+	return new;
+}
+
 static int show_regular_preop(struct expression *expr)
 {
 	int target = show_expression(expr->unop);
@@ -591,9 +730,7 @@ static int show_regular_preop(struct expression *expr)
  */
 static int show_address_gen(struct expression *expr)
 {
-	if (expr->type == EXPR_PREOP)
-		return show_expression(expr->unop);
-	return show_expression(expr->address);
+	return show_expression(expr->unop);
 }
 
 static int show_load_gen(int bits, struct expression *expr, int addr)
@@ -601,13 +738,6 @@ static int show_load_gen(int bits, struct expression *expr, int addr)
 	int new = new_pseudo();
 
 	printf("\tld.%d\t\tv%d,[v%d]\n", bits, new, addr);
-	if (expr->type == EXPR_PREOP)
-		return new;
-
-	/* bitfield load! */
-	if (expr->bitpos)
-		printf("\tshr.%d\t\tv%d,v%d,$%d\n", bits, new, new, expr->bitpos);
-	printf("\tandi.%d\t\tv%d,v%d,$%llu\n", bits, new, new, (1ULL << expr->nrbits)-1);
 	return new;
 }
 
@@ -632,6 +762,37 @@ static int show_assignment(struct expression *expr)
 	return val;
 }
 
+static int show_return_stmt(struct statement *stmt)
+{
+	struct expression *expr = stmt->ret_value;
+	struct symbol *target = stmt->ret_target;
+
+	if (expr && expr->ctype) {
+		int val = show_expression(expr);
+		int bits = expr->ctype->bit_size;
+		int addr = show_symbol_expr(target);
+		show_store_gen(bits, val, NULL, addr);
+	}
+	printf("\tret\t\t(%p)\n", target);
+	return 0;
+}
+
+static int show_initialization(struct symbol *sym, struct expression *expr)
+{
+	int val, addr, bits;
+
+	if (!expr->ctype)
+		return 0;
+
+	bits = expr->ctype->bit_size;
+	val = show_expression(expr);
+	addr = show_symbol_expr(sym);
+	// FIXME! The "target" expression is for bitfield store information.
+	// Leave it NULL, which works fine.
+	show_store_gen(bits, val, NULL, addr);
+	return 0;
+}
+
 static int show_access(struct expression *expr)
 {
 	int addr = show_address_gen(expr);
@@ -677,15 +838,18 @@ static int show_symbol_expr(struct symbol *sym)
 {
 	int new = new_pseudo();
 
+	if (sym->initializer && sym->initializer->type == EXPR_STRING)
+		return show_string_expr(sym->initializer);
+
 	if (sym->ctype.modifiers & (MOD_TOPLEVEL | MOD_EXTERN | MOD_STATIC)) {
-		printf("\tmovi.%d\t\tv%d,$%s\n", BITS_IN_POINTER, new, show_ident(sym->ident));
+		printf("\tmovi.%d\t\tv%d,$%s\n", bits_in_pointer, new, show_ident(sym->ident));
 		return new;
 	}
 	if (sym->ctype.modifiers & MOD_ADDRESSABLE) {
-		printf("\taddi.%d\t\tv%d,vFP,$%lld\n", BITS_IN_POINTER, new, sym->value);
+		printf("\taddi.%d\t\tv%d,vFP,$%lld\n", bits_in_pointer, new, sym->value);
 		return new;
 	}
-	printf("\taddi.%d\t\tv%d,vFP,$xxx\n", BITS_IN_POINTER, new);
+	printf("\taddi.%d\t\tv%d,vFP,$offsetof(%s:%p)\n", bits_in_pointer, new, show_ident(sym->ident), sym);
 	return new;
 }
 
@@ -746,17 +910,28 @@ static int show_value(struct expression *expr)
 	return new;
 }
 
+static int show_fvalue(struct expression *expr)
+{
+	int new = new_pseudo();
+	long double value = expr->fvalue;
+
+	printf("\tmovf.%d\t\tv%d,$%Lf\n", expr->ctype->bit_size, new, value);
+	return new;
+}
+
 static int show_string_expr(struct expression *expr)
 {
 	int new = new_pseudo();
 
-	printf("\tmovi.%d\t\tv%d,&%s\n", BITS_IN_POINTER, new, show_string(expr->string));
+	printf("\tmovi.%d\t\tv%d,&%s\n", bits_in_pointer, new, show_string(expr->string));
 	return new;
 }
 
-static int show_bitfield_expr(struct expression *expr)
+static int show_label_expr(struct expression *expr)
 {
-	return show_access(expr);
+	int new = new_pseudo();
+	printf("\tmovi.%d\t\tv%d,.L%p\n",bits_in_pointer, new, expr->label_symbol);
+	return new;
 }
 
 static int show_conditional_expr(struct expression *expr)
@@ -766,8 +941,6 @@ static int show_conditional_expr(struct expression *expr)
 	int false = show_expression(expr->cond_false);
 	int new = new_pseudo();
 
-	if (!true)
-		true = cond;
 	printf("[v%d]\tcmov.%d\t\tv%d,v%d,v%d\n", cond, expr->ctype->bit_size, new, true, false);
 	return new;
 }
@@ -780,10 +953,13 @@ static int show_statement_expr(struct expression *expr)
 static int show_position_expr(struct expression *expr, struct symbol *base)
 {
 	int new = show_expression(expr->init_expr);
-	struct symbol *ctype = expr->init_sym;
+	struct symbol *ctype = expr->init_expr->ctype;
+	int bit_offset;
+
+	bit_offset = ctype ? ctype->bit_offset : -1;
 
 	printf("\tinsert v%d at [%d:%d] of %s\n", new,
-		expr->init_offset, ctype->bit_offset,
+		expr->init_offset, bit_offset,
 		show_ident(base->ident));
 	return 0;
 }
@@ -793,6 +969,8 @@ static int show_initializer_expr(struct expression *expr, struct symbol *ctype)
 	struct expression *entry;
 
 	FOR_EACH_PTR(expr->expr_list, entry) {
+
+again:
 		// Nested initializers have their positions already
 		// recursively calculated - just output them too
 		if (entry->type == EXPR_INITIALIZER) {
@@ -800,15 +978,37 @@ static int show_initializer_expr(struct expression *expr, struct symbol *ctype)
 			continue;
 		}
 
-		// Ignore initializer indexes and identifiers - the
-		// evaluator has taken them into account
-		if (entry->type != EXPR_POS)
+		// Initializer indexes and identifiers should
+		// have been evaluated to EXPR_POS
+		if (entry->type == EXPR_IDENTIFIER) {
+			printf(" AT '%s':\n", show_ident(entry->expr_ident));
+			entry = entry->ident_expression;
+			goto again;
+		}
+			
+		if (entry->type == EXPR_INDEX) {
+			printf(" AT '%d..%d:\n", entry->idx_from, entry->idx_to);
+			entry = entry->idx_expression;
+			goto again;
+		}
+		if (entry->type == EXPR_POS) {
+			show_position_expr(entry, ctype);
 			continue;
-		show_position_expr(entry, ctype);
-	} END_FOR_EACH_PTR;
+		}
+		show_initialization(ctype, entry);
+	} END_FOR_EACH_PTR(entry);
 	return 0;
 }
 
+int show_symbol_expr_init(struct symbol *sym)
+{
+	struct expression *expr = sym->initializer;
+
+	if (expr)
+		show_expression(expr);
+	return show_symbol_expr(sym);
+}
+
 /*
  * Print out an expression. Return the pseudo that contains the
  * variable.
@@ -821,7 +1021,7 @@ int show_expression(struct expression *expr)
 	if (!expr->ctype) {
 		struct position *pos = &expr->pos;
 		printf("\tno type at %s:%d:%d\n",
-			input_streams[pos->stream].name,
+			stream_name(pos->stream),
 			pos->line, pos->pos);
 		return 0;
 	}
@@ -833,8 +1033,9 @@ int show_expression(struct expression *expr)
 	case EXPR_ASSIGNMENT:
 		return show_assignment(expr);
 
-	case EXPR_BINOP:
 	case EXPR_COMMA:
+		return show_comma(expr);
+	case EXPR_BINOP:
 	case EXPR_COMPARE:
 	case EXPR_LOGICAL:
 		return show_binop(expr);
@@ -846,36 +1047,45 @@ int show_expression(struct expression *expr)
 		return show_symbol_expr(expr->symbol);
 	case EXPR_DEREF:
 	case EXPR_SIZEOF:
-		warn(expr->pos, "invalid expression after evaluation");
+	case EXPR_PTRSIZEOF:
+	case EXPR_ALIGNOF:
+		warning(expr->pos, "invalid expression after evaluation");
 		return 0;
 	case EXPR_CAST:
+	case EXPR_IMPLIED_CAST:
 		return show_cast_expr(expr);
 	case EXPR_VALUE:
 		return show_value(expr);
+	case EXPR_FVALUE:
+		return show_fvalue(expr);
 	case EXPR_STRING:
 		return show_string_expr(expr);
-	case EXPR_BITFIELD:
-		return show_bitfield_expr(expr);
 	case EXPR_INITIALIZER:
 		return show_initializer_expr(expr, expr->ctype);
+	case EXPR_SELECT:
 	case EXPR_CONDITIONAL:
 		return show_conditional_expr(expr);
 	case EXPR_STATEMENT:
 		return show_statement_expr(expr);
+	case EXPR_LABEL:
+		return show_label_expr(expr);
+	case EXPR_SLICE:
+		return show_slice(expr);
 
 	// None of these should exist as direct expressions: they are only
 	// valid as sub-expressions of initializers.
 	case EXPR_POS:
-		warn(expr->pos, "unable to show plain initializer position expression");
+		warning(expr->pos, "unable to show plain initializer position expression");
 		return 0;
 	case EXPR_IDENTIFIER:
-		warn(expr->pos, "unable to show identifier expression");
+		warning(expr->pos, "unable to show identifier expression");
 		return 0;
 	case EXPR_INDEX:
-		warn(expr->pos, "unable to show index expression");
+		warning(expr->pos, "unable to show index expression");
+		return 0;
+	case EXPR_TYPE:
+		warning(expr->pos, "unable to show type expression");
 		return 0;
 	}
 	return 0;
 }
-
-
