@@ -1,8 +1,9 @@
 /*
  * This is a really stupid C tokenizer. It doesn't do any include
- * files or anything complex at all. That's the pre-processor.
+ * files or anything complex at all. That's the preprocessor.
  *
  * Copyright (C) 2003 Transmeta Corp.
+ *               2003 Linus Torvalds
  *
  *  Licensed under the Open Software License version 1.1
  */
@@ -13,9 +14,9 @@
 #include <string.h>
 #include <ctype.h>
 #include <unistd.h>
-#include <sys/stat.h>
 
 #include "lib.h"
+#include "allocate.h"
 #include "token.h"
 #include "symbol.h"
 
@@ -29,12 +30,32 @@ static int input_streams_allocated;
 
 typedef struct {
 	int fd, offset, size;
-	struct position pos;
+	int pos, line, nr;
+	int newline, whitespace;
 	struct token **tokenlist;
 	struct token *token;
 	unsigned char *buffer;
 } stream_t;
 
+const char *stream_name(int stream)
+{
+	if (stream < 0 || stream > input_stream_nr)
+		return "<bad stream>";
+	return input_streams[stream].name;
+}
+
+static struct position stream_pos(stream_t *stream)
+{
+	struct position pos;
+	pos.type = 0;
+	pos.stream = stream->nr;
+	pos.newline = stream->newline;
+	pos.whitespace = stream->whitespace;
+	pos.pos = stream->pos;
+	pos.line = stream->line;
+	pos.noexpand = 0;
+	return pos;
+}
 
 const char *show_special(int val)
 {
@@ -57,7 +78,7 @@ const char *show_ident(const struct ident *ident)
 	return buffer;
 }
 
-char *charstr(char *ptr, unsigned char c, unsigned char escape, unsigned char next)
+static char *charstr(char *ptr, unsigned char c, unsigned char escape, unsigned char next)
 {
 	if (isprint(c)) {
 		if (c == escape || c == '\\')
@@ -82,14 +103,16 @@ char *charstr(char *ptr, unsigned char c, unsigned char escape, unsigned char ne
 
 const char *show_string(const struct string *string)
 {
-	static char buffer[256];
+	static char buffer[4 * MAX_STRING + 3];
 	char *ptr;
 	int i;
 
+	if (!string->length)
+		return "<bad_string>";
 	ptr = buffer;
 	*ptr++ = '"';
 	for (i = 0; i < string->length-1; i++) {
-		const unsigned char *p = string->data + i;
+		const char *p = string->data + i;
 		ptr = charstr(ptr, p[0], '"', p[1]);
 	}
 	*ptr++ = '"';
@@ -116,21 +139,8 @@ const char *show_token(const struct token *token)
 	case TOKEN_STRING:
 		return show_string(token->string);
 
-	case TOKEN_INTEGER: {
-		const char *p = token->integer;
-		switch (*p) {
-		case 'o':	// octal
-		case 'x':	// hex
-			buffer[0] = '0';
-			strcpy(buffer+1, p+1);
-			return buffer;
-		default:
-			return p;
-		}
-	}
-
-	case TOKEN_FP:
-		return token->fp;
+	case TOKEN_NUMBER:
+		return token->number;
 
 	case TOKEN_SPECIAL:
 		return show_special(token->special);
@@ -146,11 +156,11 @@ const char *show_token(const struct token *token)
 	}
 
 	case TOKEN_STREAMBEGIN:
-		sprintf(buffer, "<beginning of '%s'>", (input_streams + token->pos.stream)->name);
+		sprintf(buffer, "<beginning of '%s'>", stream_name(token->pos.stream));
 		return buffer;
 
 	case TOKEN_STREAMEND:
-		sprintf(buffer, "<end of '%s'>", (input_streams + token->pos.stream)->name);
+		sprintf(buffer, "<end of '%s'>", stream_name(token->pos.stream));
 		return buffer;
 	
 	default:
@@ -158,7 +168,7 @@ const char *show_token(const struct token *token)
 	}
 }
 
-int init_stream(const char *name, int fd)
+int init_stream(const char *name, int fd, const char **next_path)
 {
 	int stream = input_stream_nr;
 	struct stream *current;
@@ -174,22 +184,9 @@ int init_stream(const char *name, int fd)
 	memset(current, 0, sizeof(*current));
 	current->name = name;
 	current->fd = fd;
-	current->constant = -1;	// "unknown"
-	if (fd > 0) {
-		int i;
-		struct stat st;
-
-		fstat(fd, &st);
-		current->dev = st.st_dev;
-		current->ino = st.st_ino;
-		for (i = 0; i < stream; i++) {
-			struct stream *s = input_streams + i;
-			if (s->dev == st.st_dev && s->ino == st.st_ino) {
-				if (s->constant > 0 && lookup_symbol(s->protect, NS_PREPROCESSOR))
-					return -1;
-			}
-		}
-	}
+	current->next_path = next_path;
+	current->path = NULL;
+	current->constant = CONSTANT_FILE_MAYBE;
 	input_stream_nr = stream+1;
 	return stream;
 }
@@ -197,33 +194,111 @@ int init_stream(const char *name, int fd)
 static struct token * alloc_token(stream_t *stream)
 {
 	struct token *token = __alloc_token(0);
-	token->pos = stream->pos;
+	token->pos = stream_pos(stream);
 	return token;
 }
 
-static int nextchar(stream_t *stream)
+/*
+ *  Argh...  That was surprisingly messy - handling '\r' complicates the
+ *  things a _lot_.
+ */
+static int nextchar_slow(stream_t *stream)
 {
 	int offset = stream->offset;
 	int size = stream->size;
 	int c;
+	int spliced = 0, had_cr, had_backslash, complain;
 
+restart:
+	had_cr = had_backslash = complain = 0;
+
+repeat:
 	if (offset >= size) {
 		size = read(stream->fd, stream->buffer, BUFSIZE);
 		if (size <= 0)
-			return EOF;
+			goto got_eof;
 		stream->size = size;
-		stream->offset = 0;
-		offset = 0;
+		stream->offset = offset = 0;
 	}
-	c = stream->buffer[offset];
-	stream->offset = offset + 1;
-	stream->pos.pos++;
+
+	c = stream->buffer[offset++];
+
+	if (had_cr && c != '\n')
+		complain = 1;
+
+	if (c == '\r') {
+		had_cr = 1;
+		goto repeat;
+	}
+
+	stream->pos++;
+
 	if (c == '\n') {
-		stream->pos.line++;
-		stream->pos.newline = 1;
-		stream->pos.pos = 0;
+		stream->line++;
+		stream->pos = 0;
 	}
+
+	if (!had_backslash) {
+		if (c == '\\') {
+			had_backslash = 1;
+			goto repeat;
+		}
+		if (c == '\n')
+			stream->newline = 1;
+	} else {
+		if (c == '\n') {
+			if (complain)
+				warning(stream_pos(stream), "non-ASCII data stream");
+			spliced = 1;
+			goto restart;
+		}
+		stream->pos--;
+		offset--;
+		c = '\\';
+	}
+
+out:
+	stream->offset = offset;
+	if (complain)
+		warning(stream_pos(stream), "non-ASCII data stream");
+
 	return c;
+
+got_eof:
+	if (had_backslash) {
+		c = '\\';
+		goto out;
+	}
+	if (stream->pos)
+		warning(stream_pos(stream), "no newline at end of file");
+	else if (had_cr)
+		warning(stream_pos(stream), "non-ASCII data stream");
+	else if (spliced)
+		warning(stream_pos(stream), "backslash-newline at end of file");
+	return EOF;
+}
+
+/*
+ *  We want that as light as possible while covering all normal cases.
+ *  Slow path (including the logics with line-splicing and EOF sanity
+ *  checks) is in nextchar_slow().
+ */
+static inline int nextchar(stream_t *stream)
+{
+	int offset = stream->offset;
+
+	if (offset < stream->size) {
+		int c = stream->buffer[offset++];
+		static const char special[256] = {
+			['\r'] = 1, ['\n'] = 1, ['\\'] = 1
+		};
+		if (!special[c]) {
+			stream->offset = offset;
+			stream->pos++;
+			return c;
+		}
+	}
+	return nextchar_slow(stream);
 }
 
 struct token eof_token_entry;
@@ -258,71 +333,101 @@ static void add_token(stream_t *stream)
 
 static void drop_token(stream_t *stream)
 {
-	stream->pos.newline |= stream->token->pos.newline;
-	stream->pos.whitespace |= stream->token->pos.whitespace;
+	stream->newline |= stream->token->pos.newline;
+	stream->whitespace |= stream->token->pos.whitespace;
 	stream->token = NULL;
 }
 
-static int get_base_number(unsigned int base, char **p, int next, stream_t *stream)
-{
-	char *buf = *p;
+enum {
+	Letter = 1,
+	Digit = 2,
+	Hex = 4,
+	Exp = 8,
+	Dot = 16,
+	ValidSecond = 32,
+};
 
-	*buf++ = next;
-	for (;;) {		
-		unsigned int n;
-		next = nextchar(stream);
-		n = hexval(next);
-		if (n >= base)
-			break;
-		*buf++ = next;
-	}
-	*p = buf;
-	return next;
-}
+static const long cclass[257] = {
+	['0' + 1 ... '9' + 1] = Digit | Hex,
+	['A' + 1 ... 'D' + 1] = Letter | Hex,
+	['E' + 1] = Letter | Hex | Exp,
+	['F' + 1] = Letter | Hex,
+	['G' + 1 ... 'O' + 1] = Letter,
+	['P' + 1] = Letter | Exp,
+	['Q' + 1 ... 'Z' + 1] = Letter,
+	['a' + 1 ... 'd' + 1] = Letter | Hex,
+	['e' + 1] = Letter | Hex | Exp,
+	['f' + 1] = Letter | Hex,
+	['g' + 1 ... 'o' + 1] = Letter,
+	['p' + 1] = Letter | Exp,
+	['q' + 1 ... 'z' + 1] = Letter,
+	['_' + 1] = Letter,
+	['.' + 1] = Dot | ValidSecond,
+	['=' + 1] = ValidSecond,
+	['+' + 1] = ValidSecond,
+	['-' + 1] = ValidSecond,
+	['>' + 1] = ValidSecond,
+	['<' + 1] = ValidSecond,
+	['&' + 1] = ValidSecond,
+	['|' + 1] = ValidSecond,
+	['#' + 1] = ValidSecond,
+};
 
-static int do_integer(char *buffer, int len, int next, stream_t *stream)
+/*
+ * pp-number:
+ *	digit
+ *	. digit
+ *	pp-number digit
+ *	pp-number identifier-nodigit
+ *	pp-number e sign
+ *	pp-number E sign
+ *	pp-number p sign
+ *	pp-number P sign
+ *	pp-number .
+ */
+static int get_one_number(int c, int next, stream_t *stream)
 {
-	struct token *token = stream->token;
-	void *buf;
-	
-	while (next == 'u' || next == 'U' || next == 'l' || next == 'L') {
-		buffer[len++] = next;
-		next = nextchar(stream);
-	}
-	buffer[len++] = '\0';
-	buf = __alloc_bytes(len);
-	memcpy(buf, buffer, len);
-	token_type(token) = TOKEN_INTEGER;
-	token->integer = buf;
-	add_token(stream);
-	return next;
-}
-
-static int get_one_number(int c, stream_t *stream)
-{
-	static char buffer[256];
-	int next = nextchar(stream);
-	char *p = buffer;
+	struct token *token;
+	static char buffer[4095];
+	char *p = buffer, *buf, *buffer_end = buffer + sizeof (buffer);
+	int len;
 
 	*p++ = c;
-	switch (next) {
-	case '0'...'7':
-		if (c == '0') {
-			buffer[0] = 'o';
-			next = get_base_number(8, &p, next, stream);
+	for (;;) {
+		long class =  cclass[next + 1];
+		if (!(class & (Dot | Digit | Letter)))
 			break;
-		}
-		/* fallthrough */
-	case '8'...'9':
-		next = get_base_number(10, &p, next, stream);
-		break;
-	case 'x': case 'X':
-		if (c == '0') {
-			buffer[0] = 'x';
-			next = get_base_number(16, &p, next, stream);
+		if (p != buffer_end)
+			*p++ = next;
+		next = nextchar(stream);
+		if (class & Exp) {
+			if (next == '-' || next == '+') {
+				if (p != buffer_end)
+					*p++ = next;
+				next = nextchar(stream);
+			}
 		}
 	}
-	return do_integer(buffer, p - buffer, next, stream);
+
+	if (p == buffer_end) {
+		sparse_error(stream_pos(stream), "number token exceeds %td characters",
+		      buffer_end - buffer);
+		// Pretend we saw just "1".
+		buffer[0] = '1';
+		p = buffer + 1;
+	}
+
+	*p++ = 0;
+	len = p - buffer;
+	buf = __alloc_bytes(len);
+	memcpy(buf, buffer, len);
+
+	token = stream->token;
+	token_type(token) = TOKEN_NUMBER;
+	token->number = buf;
+	add_token(stream);
+
+	return next;
 }
 
 static int escapechar(int first, int type, stream_t *stream, int *valp)
@@ -333,25 +438,46 @@ static int escapechar(int first, int type, stream_t *stream, int *valp)
 	value = first;
 
 	if (first == '\n')
-		warn(stream->pos, "Newline in string or character constant");
+		warning(stream_pos(stream), "Newline in string or character constant");
 
 	if (first == '\\' && next != EOF) {
 		value = next;
 		next = nextchar(stream);
 		if (value != type) {
 			switch (value) {
-			case 'n':
-				value = '\n';
+			case 'a':
+				value = '\a';
+				break;
+			case 'b':
+				value = '\b';
 				break;
 			case 't':
 				value = '\t';
 				break;
+			case 'n':
+				value = '\n';
+				break;
+			case 'v':
+				value = '\v';
+				break;
+			case 'f':
+				value = '\f';
+				break;
+			case 'r':
+				value = '\r';
+				break;
+			case 'e':
+				value = '\e';
+				break;
 			case '\\':
 				break;
 			case '\'':
 				break;
 			case '"':
 				break;
+			case '\n':
+				warning(stream_pos(stream), "Newline in string or character constant");
+				break;
 			case '0'...'7': {
 				int nr = 2;
 				value -= '0';
@@ -377,9 +503,9 @@ static int escapechar(int first, int type, stream_t *stream, int *valp)
 					break;
 				}
 			}
-			/* Fallthrough */
+			/* Fall through */
 			default:
-				warn(stream->pos, "Unknown escape '%c'", value);
+				warning(stream_pos(stream), "Unknown escape '%c'", value);
 			}
 		}
 		/* Mark it as escaped */
@@ -396,7 +522,7 @@ static int get_char_token(int next, stream_t *stream)
 
 	next = escapechar(next, '\'', stream, &value);
 	if (value == '\'' || next != '\'') {
-		warn(stream->pos, "Bad character constant");
+		sparse_error(stream_pos(stream), "Bad character constant");
 		drop_token(stream);
 		return next;
 	}
@@ -411,7 +537,7 @@ static int get_char_token(int next, stream_t *stream)
 
 static int get_string_token(int next, stream_t *stream)
 {
-	static char buffer[512];
+	static char buffer[MAX_STRING];
 	struct string *string;
 	struct token *token;
 	int len = 0;
@@ -422,18 +548,18 @@ static int get_string_token(int next, stream_t *stream)
 		if (val == '"')
 			break;
 		if (next == EOF) {
-			warn(stream->pos, "Enf of file in middle of string");
+			warning(stream_pos(stream), "End of file in middle of string");
 			return next;
 		}
-		if (len < sizeof(buffer)) {
+		if (len < MAX_STRING)
 			buffer[len] = val;
-			len++;
-		}
-			
+		len++;
 	}
 
-	if (len > 256)
-		warn(stream->pos, "String too long");
+	if (len > MAX_STRING) {
+		warning(stream_pos(stream), "string too long (%d bytes, %d bytes max)", len, MAX_STRING);
+		len = MAX_STRING;
+	}
 
 	string = __alloc_string(len+1);
 	memcpy(string->data, buffer, len);
@@ -465,38 +591,103 @@ static int drop_stream_eoln(stream_t *stream)
 
 static int drop_stream_comment(stream_t *stream)
 {
-	int next = nextchar(stream);
+	int newline;
+	int next;
 	drop_token(stream);
+	newline = stream->newline;
+
+	next = nextchar(stream);
 	for (;;) {
 		int curr = next;
 		if (curr == EOF) {
-			warn(stream->pos, "End of file in the middle of a comment");
+			warning(stream_pos(stream), "End of file in the middle of a comment");
 			return curr;
 		}
 		next = nextchar(stream);
 		if (curr == '*' && next == '/')
 			break;
 	}
+	stream->newline = newline;
 	return nextchar(stream);
 }
 
 unsigned char combinations[][3] = COMBINATION_STRINGS;
 
-#define NR_COMBINATIONS (sizeof(combinations)/3)
+#define NR_COMBINATIONS (SPECIAL_ARG_SEPARATOR - SPECIAL_BASE)
+
+/* hash function for two-character punctuators - all give unique values */
+#define special_hash(c0, c1) (((c0*8+c1*2)+((c0*8+c1*2)>>5))&31)
+
+/*
+ * note that we won't get false positives - special_hash(0,0) is 0 and
+ * entry 0 is filled (by +=), so all the missing ones are OK.
+ */
+static unsigned char hash_results[32][2] = {
+#define RES(c0, c1) [special_hash(c0, c1)] = {c0, c1}
+	RES('+', '='), /* 00 */
+	RES('/', '='), /* 01 */
+	RES('^', '='), /* 05 */
+	RES('&', '&'), /* 07 */
+	RES('#', '#'), /* 08 */
+	RES('<', '<'), /* 0a */
+	RES('<', '='), /* 0c */
+	RES('!', '='), /* 0e */
+	RES('%', '='), /* 0f */
+	RES('-', '-'), /* 10 */
+	RES('-', '='), /* 11 */
+	RES('-', '>'), /* 13 */
+	RES('=', '='), /* 15 */
+	RES('&', '='), /* 17 */
+	RES('*', '='), /* 18 */
+	RES('.', '.'), /* 1a */
+	RES('+', '+'), /* 1b */
+	RES('|', '='), /* 1c */
+	RES('>', '='), /* 1d */
+	RES('|', '|'), /* 1e */
+	RES('>', '>')  /* 1f */
+#undef RES
+};
+static int code[32] = {
+#define CODE(c0, c1, value) [special_hash(c0, c1)] = value
+	CODE('+', '=', SPECIAL_ADD_ASSIGN), /* 00 */
+	CODE('/', '=', SPECIAL_DIV_ASSIGN), /* 01 */
+	CODE('^', '=', SPECIAL_XOR_ASSIGN), /* 05 */
+	CODE('&', '&', SPECIAL_LOGICAL_AND), /* 07 */
+	CODE('#', '#', SPECIAL_HASHHASH), /* 08 */
+	CODE('<', '<', SPECIAL_LEFTSHIFT), /* 0a */
+	CODE('<', '=', SPECIAL_LTE), /* 0c */
+	CODE('!', '=', SPECIAL_NOTEQUAL), /* 0e */
+	CODE('%', '=', SPECIAL_MOD_ASSIGN), /* 0f */
+	CODE('-', '-', SPECIAL_DECREMENT), /* 10 */
+	CODE('-', '=', SPECIAL_SUB_ASSIGN), /* 11 */
+	CODE('-', '>', SPECIAL_DEREFERENCE), /* 13 */
+	CODE('=', '=', SPECIAL_EQUAL), /* 15 */
+	CODE('&', '=', SPECIAL_AND_ASSIGN), /* 17 */
+	CODE('*', '=', SPECIAL_MUL_ASSIGN), /* 18 */
+	CODE('.', '.', SPECIAL_DOTDOT), /* 1a */
+	CODE('+', '+', SPECIAL_INCREMENT), /* 1b */
+	CODE('|', '=', SPECIAL_OR_ASSIGN), /* 1c */
+	CODE('>', '=', SPECIAL_GTE), /* 1d */
+	CODE('|', '|', SPECIAL_LOGICAL_OR), /* 1e */
+	CODE('>', '>', SPECIAL_RIGHTSHIFT)  /* 1f */
+#undef CODE
+};
 
 static int get_one_special(int c, stream_t *stream)
 {
 	struct token *token;
-	unsigned char c1, c2, c3;
 	int next, value, i;
-	char *comb;
 
 	next = nextchar(stream);
 
 	/*
-	 * Check for strings, character constants, and comments
+	 * Check for numbers, strings, character constants, and comments
 	 */
 	switch (c) {
+	case '.':
+		if (next >= '0' && next <= '9')
+			return get_one_number(c, next, stream);
+		break;
 	case '"':
 		return get_string_token(next, stream);
 	case '\'':
@@ -512,17 +703,17 @@ static int get_one_special(int c, stream_t *stream)
 	 * Check for combinations
 	 */
 	value = c;
-	comb = combinations[0];
-	c1 = c; c2 = next; c3 = 0;
-	for (i = 0; i < NR_COMBINATIONS; i++) {
-		if (comb[0] == c1 && comb[1] == c2 && comb[2] == c3) {
-			value = i + SPECIAL_BASE;
+	if (cclass[next + 1] & ValidSecond) {
+		i = special_hash(c, next);
+		if (hash_results[i][0] == c && hash_results[i][1] == next) {
+			value = code[i];
 			next = nextchar(stream);
-			if (c3)
-				break;
-			c3 = next;
+			if (value >= SPECIAL_LEFTSHIFT &&
+			    next == "==."[value - SPECIAL_LEFTSHIFT]) {
+				value += 3;
+				next = nextchar(stream);
+			}
 		}
-		comb += 3;
 	}
 
 	/* Pass it on.. */
@@ -533,7 +724,7 @@ static int get_one_special(int c, stream_t *stream)
 	return next;
 }
 
-#define IDENT_HASH_BITS (10)
+#define IDENT_HASH_BITS (13)
 #define IDENT_HASH_SIZE (1<<IDENT_HASH_BITS)
 #define IDENT_HASH_MASK (IDENT_HASH_SIZE-1)
 
@@ -542,7 +733,7 @@ static int get_one_special(int c, stream_t *stream)
 #define ident_hash_end(hash)		((((hash) >> IDENT_HASH_BITS) + (hash)) & IDENT_HASH_MASK)
 
 static struct ident *hash_table[IDENT_HASH_SIZE];
-int ident_hit, ident_miss;
+static int ident_hit, ident_miss, idents;
 
 void show_identifier_stats(void)
 {
@@ -579,6 +770,7 @@ static struct ident *alloc_ident(const char *name, int len)
 	struct ident *ident = __alloc_ident(len);
 	ident->symbols = NULL;
 	ident->len = len;
+	ident->tainted = 0;
 	memcpy(ident->name, name, len);
 	return ident;
 }
@@ -594,17 +786,34 @@ static struct ident * insert_hash(struct ident *ident, unsigned long hash)
 static struct ident *create_hashed_ident(const char *name, int len, unsigned long hash)
 {
 	struct ident *ident;
+	struct ident **p;
 
-	ident = hash_table[hash];
-	while (ident) {
-		if (ident->len == len && !memcmp(ident->name, name, len)) {
+	p = &hash_table[hash];
+	while ((ident = *p) != NULL) {
+		if (ident->len == (unsigned char) len) {
+			const char *n = name;
+			const char *m = ident->name;
+			int l = len;
+			do {
+				if (*n != *m)
+					goto next;
+				n++;
+				m++;
+			} while (--l);
+				
 			ident_hit++;
 			return ident;
 		}
-		ident = ident->next;
+next:
+		//misses++;
+		p = &ident->next;
 	}
-
-	return insert_hash(alloc_ident(name, len), hash);
+	ident = alloc_ident(name, len);
+	*p = ident;
+	ident->next = NULL;
+	ident_miss++;
+	idents++;
+	return ident;
 }
 
 static unsigned long hash_name(const char *name, int len)
@@ -655,19 +864,13 @@ static int get_one_identifier(int c, stream_t *stream)
 	buf[0] = c;
 	for (;;) {
 		next = nextchar(stream);
-		switch (next) {
-		case '0'...'9':
-		case 'a'...'z':
-		case 'A'...'Z':
-		case '_':
-			if (len < sizeof(buf)) {
-				hash = ident_hash_add(hash, next);
-				buf[len] = next;
-				len++;
-			}
-			continue;
-		}
-		break;
+		if (!(cclass[next + 1] & (Letter | Digit)))
+			break;
+		if (len >= sizeof(buf))
+			break;
+		hash = ident_hash_add(hash, next);
+		buf[len] = next;
+		len++;
 	};
 	hash = ident_hash_end(hash);
 
@@ -683,16 +886,12 @@ static int get_one_identifier(int c, stream_t *stream)
 
 static int get_one_token(int c, stream_t *stream)
 {
-	switch (c) {
-	case '0'...'9':
-		return get_one_number(c, stream);
-	case 'a'...'z':
-	case 'A'...'Z':
-	case '_':
+	long class = cclass[c + 1];
+	if (class & Digit)
+		return get_one_number(c, nextchar(stream), stream);
+	if (class & Letter)
 		return get_one_identifier(c, stream);
-	default:
-		return get_one_special(c, stream);
-	}	
+	return get_one_special(c, stream);
 }
 
 static struct token *setup_stream(stream_t *stream, int idx, int fd,
@@ -700,11 +899,11 @@ static struct token *setup_stream(stream_t *stream, int idx, int fd,
 {
 	struct token *begin;
 
-	stream->pos.stream = idx;
-	stream->pos.line = 1;
-	stream->pos.newline = 1;
-	stream->pos.whitespace = 0;
-	stream->pos.pos = 0;
+	stream->nr = idx;
+	stream->line = 1;
+	stream->newline = 1;
+	stream->whitespace = 0;
+	stream->pos = 0;
 
 	stream->token = NULL;
 	stream->fd = fd;
@@ -722,27 +921,21 @@ static void tokenize_stream(stream_t *stream, struct token *endtoken)
 {
 	int c = nextchar(stream);
 	while (c != EOF) {
-		if (c == '\\') {
-			c = nextchar(stream);
-			stream->pos.newline = 0;
-			stream->pos.whitespace = 1;
-			continue;
-		}
 		if (!isspace(c)) {
 			struct token *token = alloc_token(stream);
 			stream->token = token;
-			stream->pos.newline = 0;
-			stream->pos.whitespace = 0;
+			stream->newline = 0;
+			stream->whitespace = 0;
 			c = get_one_token(c, stream);
 			continue;
 		}
-		stream->pos.whitespace = 1;
+		stream->whitespace = 1;
 		c = nextchar(stream);
 	}
 	mark_eof(stream, endtoken);
 }
 
-struct token * tokenize_buffer(unsigned char *buffer, unsigned long size, struct token *endtoken)
+struct token * tokenize_buffer(void *buffer, unsigned long size, struct token *endtoken)
 {
 	stream_t stream;
 	struct token *begin;
@@ -752,16 +945,18 @@ struct token * tokenize_buffer(unsigned char *buffer, unsigned long size, struct
 	return begin;
 }
 
-struct token * tokenize(const char *name, int fd, struct token *endtoken)
+struct token * tokenize(const char *name, int fd, struct token *endtoken, const char **next_path)
 {
 	struct token *begin;
 	stream_t stream;
 	unsigned char buffer[BUFSIZE];
 	int idx;
 
-	idx = init_stream(name, fd);
-	if (idx < 0)
+	idx = init_stream(name, fd, next_path);
+	if (idx < 0) {
+		// info(endtoken->pos, "File %s is const", name);
 		return endtoken;
+	}
 
 	begin = setup_stream(&stream, idx, fd, buffer, 0);
 	tokenize_stream(&stream, endtoken);
